from mint import utils
from mint import exceptions
from django.contrib.auth.models import User as AuthUser


class BaseMixin(object):
    def _filter(self, queryset):
        return queryset

    def _pre_return(self, model, out):
        pass

    def _pre_save(self, model):
        pass

    def _post_save(self, model):
        pass

    def _can_delete(self):
        return False

    def _iterate_and_serialize(self, model):
        out = []
        for m in model:
            out.append(self._serialize(m))
        return out

    def _get_serializer_fields(self):
        return []

    def _serialize(self, model):
        fields = self._get_serializer_fields()
        me = self.serializer(model, fields).pack()
        self._pre_return(model, me)
        return me

    def _get_name(self):
        return self.__class__.__name__

    def _format_plural(self, out, name=None):
        if not name:
            name = utils.camel_to_underscore(self._get_name())
        if name[-1:] != "s":
            name = "%ss" % name
        return {name: out}

    def _format_singular(self, out, name=None):
        if not name:
            name = utils.camel_to_underscore(self._get_name())
        if name[-1:] == "s":
            name = name[:-1]
        return {name: out}


class RootMixin(BaseMixin):
    def _get_root(self):
        self.model = self._filter(self.model)
        return self._format_plural(self._iterate_and_serialize(self.model))

    def _post_root(self):
        m = self.original_model()
        for key, value in self.args.items():
            if key == 'id':
                continue
            if hasattr(m, key):
                my_type = m._meta.get_field(key).get_internal_type()
                if my_type in ('DateTimeField', 'DateField'):
                    if value:
                        value = utils.datetime_from_string(value)
                setattr(m, key, value)
        self._pre_save(m)
        m.save()
        self._post_save(m)
        return self._format_singular(self._serialize(m))

    def _put_root(self):
        raise exceptions.HttpNotAllowed("Cannot PUT on root context")

    def _delete_root(self):
        raise exceptions.HttpNotAllowed("Cannot DELETE on root context")

    def _root_methods(self):
        return {
            'GET': self._get_root,
            'POST': self._post_root,
            'PUT': self._put_root,
            'DELETE': self._delete_root,
        }


class IDMixin(BaseMixin):
    def _get_by_id(self):
        return self._format_singular(self._serialize(self.model))

    def _post_by_id(self):
        raise exceptions.HttpNotAllowed("Cannot POST by ID")

    def _put_by_id(self):
        for key, value in self.args.items():
            if key == 'id':
                continue
            if hasattr(self.model, key):
                my_type = self.model._meta.get_field(key).get_internal_type()
                if my_type in ('DateTimeField', 'DateField'):
                    value = utils.datetime_from_string(value)
                setattr(self.model, key, value)
        self._pre_save(self.model)
        self.model.save()
        self._post_save(self.model)
        return self._format_singular(self._serialize(self.model))

    def _delete_by_id(self):
        if not self._can_delete():
            raise exceptions.HttpUnauthorized("You cannot delete this object")
        self.model.delete()
        return True

    def _id_methods(self):
        return {
            'GET': self._get_by_id,
            'POST': self._post_by_id,
            'PUT': self._put_by_id,
            'DELETE': self._delete_by_id,
        }


class ManyToManyMixin(BaseMixin):
    def _get_m2m_field(self, field):
        model = getattr(self.model, field).all()
        return self._format_plural(self._iterate_and_serialize(model), name=field)

    def _post_m2m_field(self, field):
        instance = getattr(self.model, field)
        m = instance.model()
        if isinstance(m, AuthUser):
            if not hasattr(self, '_create_user'):
                raise AttributeError("Trying to create new user but controller does not implement 'create_user' "
                                     "method.")
            m = getattr(self, '_create_user')()
            instance.add(m)
            return self._format_singular(self._serialize(m), name=field)
        else:
            for key, value in self.args.items():
                if key == 'id':
                    continue
                if hasattr(m, key):
                    my_type = m._meta.get_field(key).get_internal_type()
                    if my_type in ('DateTimeField', 'DateField'):
                        value = utils.datetime_from_string(value)
                    setattr(m, key, value)
            self._pre_save(m)
            m.save()
            instance.add(m)
            self._post_save(m)
            return self._format_singular(self._serialize(m), name=field)

    def _put_m2m_field(self, field):
        raise exceptions.HttpNotAllowed("Cannot PUT on a Many to Many Field")

    def _delete_m2m_field(self, field):
        raise exceptions.HttpNotAllowed("Cannot DELETE on a Many to Many Field")

    def _m2m_methods(self):
        return {
            'GET': self._get_m2m_field,
            'POST': self._post_m2m_field,
            'PUT': self._put_m2m_field,
            'DELETE': self._delete_m2m_field,
        }


from django.conf import settings
from django.core import serializers
from django.db.models.fields import FieldDoesNotExist
from mint import utils


class Serializer(object):
    def __init__(self, model, include=[]):
        self.model = model
        self.include = include

    def pack(self):
        if len(self.include) > 0:
            fields = serializers.serialize('python', (self.model, ), fields=self.include)[0]['fields']
        else:
            fields = serializers.serialize('python', (self.model, ))[0]['fields']
        for name, value in fields.items():
            try:
                my_type = self.model._meta.get_field(name).get_internal_type()
                if my_type == 'DateTimeField':
                    fields[name] = utils.string_from_datetime(value)
                elif my_type == 'DateField':
                    fields[name] = utils.string_from_datetime(value)
                elif my_type == 'TimeField':
                    fields[name] = utils.string_from_time(value)
                elif my_type == 'DecimalField':
                    fields[name] = float(value)
                elif my_type == 'FileField':
                    fields[name] = "%s%s" % (settings.MEDIA_URL, value)
                elif my_type in ('ForeignKey',):
                    key = "%s_id" % name
                    fields[key] = value
                    del fields[name]
            except FieldDoesNotExist:
                pass
        fields['id'] = self.model.id
        fields['model'] = utils.camel_to_underscore(self.model.__class__.__name__)
        return fields


# Base class
class HttpError(Exception):
    pass


# Bad Request
class HttpBadRequest(HttpError):
    pass


# Unauthorized
class HttpUnauthorized(HttpError):
    pass


# Forbidden
class HttpForbidden(HttpError):
    pass


# Not Found
class HttpNotFound(HttpError):
    pass


# Method not allowed
class HttpNotAllowed(HttpError):
    pass


class InvalidContext(Exception):
    pass


class InvalidRelation(Exception):
    pass


class ParameterError(HttpBadRequest):
    pass


import json
from mint import utils
from mint import exceptions
from django import http
from django.conf import settings
import traceback
import logging


class HttpResponseUnauthorized(http.HttpResponse):
    status_code = 403


class Coordinator(object):
    controllers = []

    def __init__(self, request, ctx, idx=None, action=None):
        self.request = request
        self.ctx = utils.underscore_to_camel(ctx)
        self.id = idx
        self.action = action
        self.related = action
        self.controller = None
        self.related_controller = None

    def open(self):
        for ctx in self.controllers:
            if ctx.__name__ == self.ctx:
                self.controller = ctx(self.request)
                return
        raise exceptions.InvalidContext('This controller does not exist: %s' % self.ctx)

    def run(self):
        try:
            return self.run_inner()
        except exceptions.HttpError as exc:
            if isinstance(exc, exceptions.HttpBadRequest):
                return self.return_error(exc, response=http.HttpResponseBadRequest)
            elif isinstance(exc, exceptions.HttpForbidden):
                return self.return_error(exc, response=http.HttpResponseForbidden)
            elif isinstance(exc, exceptions.HttpNotAllowed):
                return self.return_error(exc, response=http.HttpResponseNotAllowed)
            elif isinstance(exc, exceptions.HttpNotFound):
                return self.return_error(exc, response=http.HttpResponseNotFound)
            elif isinstance(exc, exceptions.HttpUnauthorized):
                return self.return_error(exc, response=HttpResponseUnauthorized)
            elif isinstance(exc, exceptions.HttpError):
                self.report_exception()
                return self.return_error(exc, response=http.HttpResponseServerError)
        except Exception as exc:
            self.report_exception()
            return self.return_error(exc, response=http.HttpResponseServerError)

    def run_inner(self):
        self.open()
        if self.id and (self.related or self.action):
            if self.find_related():
                return self.exec_related()
            elif self.find_m2m_field():
                return self.exec_m2m()
            elif self.find_action():
                return self.exec_action()
            else:
                raise exceptions.HttpBadRequest("No such action or related model: %s" %
                                                self.related if self.related is not None else self.action)
        elif self.id:
            return self.exec_id()
        elif not self.id and self.action:
            if self.find_action():
                return self.exec_action()
            else:
                raise exceptions.HttpBadRequest("No such action: %s" % self.action)
        else:
            return self.exec_root()

    def find_related(self):
        if len(self.controller.related) > 0:
            for ctx in self.controller.related:
                if ctx.__name__ == utils.underscore_to_camel(self.related):
                    self.related_controller = ctx(self.request)
                    return True
            return False
        return False

    def find_m2m_field(self):
        self.controller.open_model(self.id)
        if hasattr(self.controller.model, self.related):
            return True
        return False

    def find_action(self):
        return self.controller.has_action(self.action)

    def exec_related(self):
        self.controller.open_model(self.id)
        self.related_controller.open_model()
        filter = {
            utils.camel_to_underscore(self.controller.__class__.__name__): self.controller.model
        }
        self.related_controller.model.filter(**filter)
        return self.return_success(self.related_controller.exec_root())

    def exec_m2m(self):
        return self.return_success(self.controller.exec_m2m(self.id, self.related))

    def exec_action(self):
        return self.return_success(self.controller.exec_action(self.action, self.id))

    def exec_id(self):
        return self.return_success(self.controller.exec_id(self.id))

    def exec_root(self):
        self.controller.open_model()
        return self.return_success(self.controller.exec_root())

    def return_success(self, out, response=http.HttpResponse):
        if isinstance(out, dict):
            out['result'] = True
        return self.format(out, response=response)

    def return_error(self, exc=None, error=None, message=None, response=http.HttpResponse):
        if exc:
            return self.format({
                'result': False,
                'message': str(exc),
                'error': exc.__class__.__name__,
                'traceback': traceback.format_exc() if settings.DEBUG else '',
            }, response=response)
        elif error and message:
            return self.format({
                'result': False,
                'message': message,
                'error': error
            }, response=response)
        else:
            return self.format({
                'result': False,
                'message': 'Error handler not correctly invoked',
                'error': 'Exception'
            }, response=response)

    def format(self, out, response=http.HttpResponse):
        if out is True:
            return self.format({'result': True}, response=response)
        else:
            return response(json.dumps(out), content_type='application/json')

    def report_exception(self):
        logger = logging.getLogger('django.request')
        logger.exception(
            "Internal Server Error during API call: %s" % self.request.path,
            extra={'status_code': 500, 'request': self.request}
        )


from django.conf.urls import patterns, url
from mint import views


urlpatterns = [
    url(r'^([a-z_]+)/([0-9]+)/([a-z_]+)/?$', views.with_id),
    url(r'^([a-z_]+)/([0-9]+)/?$', views.with_id),
    url(r'^([a-z_]+)/([a-z_]+)/?$', views.without_id),
    url(r'^([a-z_]+)/?$', views.without_id),
]


from mint import exceptions


def get(func):
    def wrapper(self):
        if self.request.method != "GET":
            raise exceptions.HttpNotAllowed("This action needs to be a HTTP GET.")
        return func(self)
    return wrapper


def post(func):
    def wrapper(self):
        if self.request.method != "POST":
            raise exceptions.HttpNotAllowed("This action needs to be a HTTP POST.")
        return func(self)
    return wrapper


def put(func):
    def wrapper(self):
        if self.request.method != "PUT":
            raise exceptions.HttpNotAllowed("This action needs to be a HTTP PUT.")
        return func(self)
    return wrapper


def delete(func):
    def wrapper(self):
        if self.request.method != "DELETE":
            raise exceptions.HttpNotAllowed("This action needs to be a HTTP DELETE.")
        return func(self)
    return wrapper


def requires_id(func):
    def wrapper(self):
        if hasattr(self, 'id') and self.id is None:
            raise exceptions.ParameterError("This action requires an ID")
        return func(self)
    return wrapper


def requires(**kwargs):
    def decorator(function):
        def wrapper(self):
            for arg, typex in kwargs.items():
                if arg not in self.args.keys():
                    raise exceptions.ParameterError("This action requires argument '%s' but it was not passed." % arg)
                if not isinstance(self.args[arg], typex):
                    try:
                        typex(self.args[arg])
                    except ValueError:
                        raise exceptions.ParameterError("Could not convert argument '%s' to a '%s'" %
                                                        (arg, typex.__class__.__name__))
            return function(self)
        return wrapper
    return decorator


from django.conf import settings


def with_id(request, ctx, idx, action=None):
    coordinator = get_coordinator()
    api = coordinator(request, ctx, idx=idx, action=action)
    return api.run()


def without_id(request, ctx, action=None):
    coordinator = get_coordinator()
    api = coordinator(request, ctx, action=action)
    return api.run()


def get_coordinator():
    split = settings.REST_COORDINATOR_PATH.split(".")
    path = ".".join(split[:-1])
    mod = __import__(path, fromlist=[split[-1:][0]])
    return getattr(mod, split[-1:][0])


import dateutil.parser


def datetime_from_string(string):
    return dateutil.parser.parse(string)


def string_from_datetime(date):
    if date is None:
        return ""
    if isinstance(date, str):
        date = dateutil.parser.parse(date)
    return date.strftime("%Y-%m-%dT%H:%M:%S")


def time_from_string(string):
    return dateutil.parser.parse(string)


def string_from_time(time):
    if time is None:
        return ""
    return time.strftime("%H:%M:%S+0000")


def underscore_to_camel(string):
    name = ""
    for s in string.split("_"):
        name += s[0].upper() + s[1:]
    return name


def camel_to_underscore(string):
    name = ""
    i = 0
    for s in string:
        if s.isupper() and i > 0:
            name += "_"
        name += s.lower()
        i += 1
    return name